#lang scheme

(require scheme/cmdline
         "board.ss"
         "find.ss"
         "test.ss")

(provide play-leech)

;; receive : input-port symbol -> any
;;  Receives a particular expected message, and returns
;;  the data produced by `read'
(define (receive r expected-msg)
  (let ([got-msg (read r)])
    (if (eq? got-msg expected-msg)
        (read r)
        (error 'receive "expected ~s; got ~s" expected-msg got-msg))))

;; should-roll? : state num -> boolean
;;  Determines whether we must not roll on this turn
(define (should-roll? state me)
  (define (none? v) (equal? v #hash()))
  (let-values ([(roll-destss nonroll-dests)
                (find-all-possible state me)])
    (cond
      ;; No way to move if we roll? Then don't roll.
      [(andmap none? roll-destss) #f]
      ;; Otherwise, just roll:
      [else #t])))

(define (move state me roll)
  (let ([dests (find-dests state me roll)])
    (if (zero? (hash-count dests))
        ;; We can't move
        (list-ref state me)
        ;; Pick randomly:
        (random-ref
         (for/list ([pos (in-hash-keys dests)]) pos)))))

;; say : output-port any -> void
;;  Writes the given reply
(define (say w what)
  (fprintf w "~a\n" what)
  (flush-output w))

;; handle-other-turn : input-port output-port num list -> void
;;  Handles a `moved' message, supplying a random card
;;  from our hand as needed to disprove a suggestion
(define (handle-other-turn r w me hand)
  (let-values ([(state who suggestion who-disproved)
                (apply values (receive r 'moved))])
    (cond
      [(equal? me who-disproved)
       ;; Need to pick a card from our hand that
       ;;  disproves the suggestion:
       (say w (random-ref
               (for/list ([t (in-list suggestion)]
                          #:when (member t hand))
                 t)))
       #f]
      [(and (list? suggestion)
            (eq? who-disproved '-))
       ;; Steal this suggestion, since no one disproved it:
       suggestion]
      [else #f])))

(define num-others (sub1 (length all-players)))

(define (play-leech host port name)
  (let-values ([(r w) (tcp-connect host port)])
    ;; Register our name:
    (say w name)
    ;; Get our position:
    (let-values ([(me hand) (apply values (receive r 'hello))])
      ;; Loop:
      (let loop ([turns-before me])
        ;; Wait until our turn, watching for an accusation idea:
        (let ([idea (for/fold ([idea #f]) ([i (in-range turns-before)])
                      (or (handle-other-turn r w me hand)
                          idea))])
          ;; Our turn:
          (let ([state (receive r 'turn)])
            (cond
              [idea
               ;; Try to steal the game:
               (say w 'accuse)
               (receive r 'accuse)
               (say w idea)
               (receive r 'wrong) ;; OOPS!
               ;; We lost, but continue playing for others:
               (let loop () 
                 (handle-other-turn r w me hand)
                 (loop))]
              [else
               ;; Normal turn:
               (say w (if (should-roll? state me)
                          'roll
                          'no-roll))
               (let* ([roll (receive r 'rolled)]
                      [dest (move state me roll)])
                 (say w dest)
                 (cond
                   [(and (symbol? (car dest))
                         (not (equal? dest (list-ref state me))))
                    ;; We entered a room, so we have to make a suggestion:
                    (let ([here (receive r 'suggest)])
                      (say w (list 0 'rope here))
                      (receive r 'counter)
                      (loop num-others))]
                   [else
                    ;; not in a room:
                    (loop num-others)]))])))))))
