#lang scheme

(require "board.ss"
         "find.ss"
         "validate.ss"
         "test.ss")

(provide serve)

;; Randomly pick the answer to the mystery:
(define murder-player (random-ref all-players))
(define murder-weapon (random-ref all-weapons))
(define murder-room (random-ref all-rooms))

;; A player is
;;  (make-player input-port output-port string boolean)
(define-struct player (r w number name cards out?))

;; shuffle : num list -> list
;;  Shuffles the list `c' times
(define (shuffle c l)
  (if (zero? c)
      l
      (shuffle
       (sub1 c)
       (let ([n (quotient (length l) 2)])
         (let loop ([a (take l n)][b (drop l n)])
           (cond
             [(null? a) b]
             [(null? b) a]
             [(zero? (random 2))
              (cons (car a) (loop (cdr a) b))]
             [else
              (cons (car b) (loop a (cdr b)))]))))))

;; every-nth list positive-integer -> list
;;  Produces a list with every `n'th element of the given
;;  one, including with the first element
(define (every-nth l n)
  (for/list ([v (in-list l)]
             [i (in-naturals)]
             #:when (zero? (modulo i n)))
    v))

;; get-players: tcp-listener -> list-of-player
(define (get-players l)
  (define names '("Scarlet" "Mustard" "White" "Green" "Peacock" "Plum"))
  (define cards (shuffle
                 7
                 (append (remove murder-player all-players)
                         (remove murder-weapon all-weapons)
                         (remove murder-room all-rooms))))
  ;; Accept players in random order, then sort back to order:
  (sort
   (for/list ([i (in-list (shuffle 7 (for/list ([i (in-range 6)]) i)))])
     (let ([name (list-ref names i)])
       (let-values ([(r w) (tcp-accept l)])
         (let* ([hand (every-nth (drop cards i) 6)]
                [p (make-player r w 
                                i
                                (string-append (read-line r) " " name)
                                hand
                                #f)])
           (tell p 'hello (list i hand))
           p))))
   < #:key player-number))

;; tell : player symbol string -> void
;;  Sends a message to a player
(define (tell p what msg)
  (fprintf (player-w p) (format "~a\n~a\n" what msg))
  (flush-output (player-w p)))

;; check-must-roll : state nat -> string-or-#f
;;  Returns a required roll command, if any
(define (check-must-roll state me)
  (define (none? v) (equal? v #hash()))
  (let-values ([(roll-destss  nonroll-dests)
                (find-all-possible state me)])
    (cond
     [(none? nonroll-dests)
      (if (andmap none? roll-destss) 
          #f
          "roll")]
     [(andmap none? roll-destss) "no-roll"]
     [else #f])))

;; get-turn-action : string-or-#f -> number-or-#f
;;  #f result => accusation
;;  0 result => no roll
;;  otherwise => roll
(define (get-turn-action p must-str)
  (let ([cmd (read-line (player-r p))])
    (cond
      [(equal? cmd "accuse")
       #f]
      [(and must-str (not (equal? must-str cmd)))
       (error 'roll "bad roll choice: ~s" cmd)]
      [(equal? cmd "roll")
       (+ (add1 (random 6)) (add1 (random 6)))]
      [(equal? cmd "no-roll")
       0]
      [else (error 'roll "bad roll replay: ~s" cmd)])))

;; get-turn-dest : player num state (string -> void)
;;                 -> dest state
;;  Accepts a move message from a client, validates it,
;;  and produces the destination along with updated stated.
;;  Also reports a message via `msg!'.
(define (get-turn-dest p roll state msg!)
  (let* ([cmd (read-line (player-r p))]
         [dest (read (open-input-string cmd))])
    ;; Check the destination format:
    (unless (and (list? dest)
                 (= 2 (length dest))
                 (or (and (number? (car dest))
                          (number? (cadr dest)))
                     (and (room? (car dest))
                          (memq (cadr dest) '(S V)))))
      (error 'move "bad move reply: ~s" cmd))
    (msg! (format "~a ~a; moved to ~a"
                  (player-name p)
                  (if (zero? roll)
                      "didn't roll"
                      (format "rolled ~a" roll))
                  dest))
    (let ([new-state (append (take state (player-number p))
                             (list dest)
                             (drop state (add1 (player-number p))))])
      (unless (valid-move? state (player-number p) roll new-state)
        (error 'move "illegal move: ~s" cmd))
      (values dest new-state))))

;; get-turn-suggestion* : string player room-or-#f (string -> void)
;;                        -> num weapon room
;;  Accepts a suggestion or accusation from a client, validates it,
;;  and retrsun the components of the suggestion. If `in-room'
;;  is not #f, ensures that the suggested room matches (while #f
;;  is suitable for a final accusation, which can be from anywhere).
;;  Also reports a message via `msg!'.
(define (get-turn-suggestion* what p in-room msg!)
  (let* ([cmd (read-line (player-r p))]
         [sus (read (open-input-string cmd))])
    ;; Check the suggestion format:
    (unless (and (list? sus)
                 (= 3 (length sus))
                 (and (member (car sus) all-players)
                      (member (cadr sus) all-weapons)
                      (member (caddr sus) all-rooms)
                      (or (not in-room)
                          (eq? (caddr sus) in-room))))
      (error 'move "bad ~a reply: ~s" what cmd))
    (msg! (format "~as: ~a" what sus))
    (values (car sus) (cadr sus) (caddr sus))))

;; get-turn-suggestion : player room (string -> void)
;;                       -> num weapon room
;; Accepts a suggestion from a client and returns the
;; components.
(define (get-turn-suggestion p in-room msg!)
  (get-turn-suggestion* "suggest" p in-room msg!))

;; get-turn-accusation : player room (string -> void)
;;                       -> num weapon room
;; Accepts an accusation from a client and returns the
;; components.
(define (get-turn-accusation p msg!)
  (get-turn-suggestion* "accuse" p #f msg!))

;; confirm-or-disprove : num weapon room
;;                       -> num-or-weapon-or-room-or-#f
;;  Returns #f if the given combination matches the
;;  murder, otherwise picks one of the wrong parts
;;  randomly.
(define (confirm-or-disprove who what where)
  (let ([bads (append
               (if (equal? who murder-player)
                   null
                   (list who))
               (if (equal? what murder-weapon)
                   null
                   (list what))
               (if (equal? where murder-room)
                   null
                   (list where)))])
    (if (null? bads)
        #f
        (random-ref bads))))

;; move-suspect : state num room -> state
;;  Move player due to suggestion
(define (move-suspect state who where)
  (let ([dest (list-ref state who)])
    (if (eq? where (car dest))
        ;; player is already in the room
        state
        ;; move the player:
        (append (take state who)
                (list (list where 'S))
                (drop state (add1 who))))))
(test (move-suspect '((17 24) (24 17) (15 0) (10 0) (0 6) (0 19))
                    2
                    'C)
      '((17 24) (24 17) (C S) (10 0) (0 6) (0 19)))
(test (move-suspect '((17 24) (24 17) (C V) (10 0) (0 6) (0 19))
                    2
                    'C)
      '((17 24) (24 17) (C V) (10 0) (0 6) (0 19)))

;; find-first-disprove : list-of-players num weapon room 
;;                       -> player-or-#f
(define (find-first-disprove players who what where)
  (for/or ([p (in-list players)])
    (and (for/or ([t (list who what where)])
           (memq t (player-cards p)))
         p)))

;; choose-counter : player player state num weapon room 
;;                  (string -> any)
;;                  -> any
;;  For the `who'-`what'-`where' suggestion by `p', asks
;;  `dp' which of its cards it would like to show (in case
;;  there's  choice); returns the choice.
(define (choose-counter dp p state who what where msg!)
  (let ([choices (for/list ([t (list who what where)]
                            #:when (member t (player-cards dp)))
                   t)])
    (tell dp 'moved (list state
                          (player-number p)
                          (list who what where)
                          (player-number dp)))
    (let* ([cmd (read-line (player-r dp))]
           [t (read (open-input-string cmd))])
      (unless (memq t choices)
        (error 'move "bad disprove response: ~s" cmd))
      (msg! (format "~a disproves: ~a" (player-number dp) t))
      t)))

;; broadcast-move : state list-of-player player
;;                  sym-or-list player-or-sym 
;;                  -> void
(define (broadcast-move state players mp suggestion dp)
  (for ([p (in-list players)])
    (unless (or (eq? p mp) (eq? p dp))
      (tell p 'moved (list state
                           (player-number mp)
                           suggestion
                           (if (player? dp)
                               (player-number dp)
                               dp))))))

(define (serve port msg! state!)
  (define l (tcp-listen port 5 #t))
  (lambda ()
    (let ([state '((17 24) (24 17) (15 0) (10 0) (0 6) (0 19))])
      ;; Loop until someone wins. After we let the first player
      ;; move, shift the player record to the end of the list.
      (let loop ([state state]
                 [players (get-players l)])
        (let ([p (car players)]
              [next (lambda (new-state p)
                      (lambda ()
                        (loop new-state (append (cdr players) (list p)))))])
          ((with-handlers ([exn? (lambda (e)
                                   (raise-user-error 
                                    (format
                                     "during player ~s: ~a"
                                     (player-name p)
                                     (exn-message e))))])
             ;; Update display:
             (state! (format "Turn: ~a (~a)\n" 
                             (player-name p) (player-number p))
                     state)
             (cond
               [(player-out? p)
                ;; Player is out of the game due to a bad accusation
                (msg! (format "~a skipped" (player-name p)))
                (broadcast-move state players p '- '-)
                (next state p)]
               [else
                ;; Tell player to move:
                (tell p 'turn state)
                ;; Get roll request from player:
                (let ([roll (get-turn-action
                             p
                             (check-must-roll state (player-number p)))])
                  (cond
                    [roll
                     ;; Normal move. Report back roll:
                     (tell p 'rolled roll)
                     ;; Get back player's destination: 
                     (let-values ([(dest new-state) 
                                   (get-turn-dest p roll state msg!)]
                                  [(src) (list-ref state (player-number p))])
                       ;; Can name a suspect?
                       (cond
                         [(and (symbol? (car dest))
                               (not (equal? dest src)))
                          ;; Get suggestion
                          (tell p 'suggest (car dest))
                          (let-values ([(who what where) 
                                        (get-turn-suggestion p (car dest) msg!)])
                            (let ([new-state (move-suspect new-state who where)]
                                  [dp (find-first-disprove (cdr players)
                                                           who what where)])
                              (let ([counter (and dp
                                                  (choose-counter dp p new-state 
                                                                  who what where
                                                                  msg!))])
                                (broadcast-move new-state players p 
                                                (list who what where) (or dp '-))
                                (tell p 'counter (list (or counter "-") 
                                                       (if dp
                                                           (player-number dp)
                                                           '-)))
                                (next new-state p))))]
                         [else
                          ;; No suggestion possible
                          (broadcast-move new-state players p '- '-)
                          (next new-state p)]))]
                    [else
                     ;; Player is making an accusation in an attempt to win the game
                     (tell p 'accuse "?")
                     (let-values ([(who what where) (get-turn-accusation p msg!)])
                       (if (not (confirm-or-disprove who what where))
                           ;; Right: end game with winner message
                           (lambda ()
                             (msg! (format "~a wins!" (player-name p))))
                           ;; Wrong: player cannot move anymore
                           (begin
                             (msg! (format "~a loses!" (player-name p)))
                             (broadcast-move state players p 
                                             (list who what where) '!)
                             (tell p 'wrong "!")
                             (if (andmap player-out? (cdr players))
                                 (lambda () (msg! "everyone lost"))
                                 (let ([p (struct-copy player p [out? #t])])                               
                                   (next state p))))))]))]))))))))
